r"""Contain the base class to implement an output."""

from __future__ import annotations

__all__ = ["BaseOutput"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig

    from arkas.content.base import BaseContentGenerator
    from arkas.evaluator2.base import BaseEvaluator
    from arkas.plotter.base import BasePlotter


class BaseOutput(ABC):
    r"""Define the base class to implement an output.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.output import AccuracyOutput
    >>> from arkas.state import AccuracyState
    >>> output = AccuracyOutput(
    ...     AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> output
    AccuracyOutput(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    @abstractmethod
    def compute(self) -> BaseOutput:
        r"""Compute the results and return a new ouptut.

        Returns:
            A new ouptut with the computed results.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> output = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output
        AccuracyOutput(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )
        >>> out = output.compute()
        >>> out
        Output(
          (content): ContentGenerator()
          (evaluator): Evaluator(count=5)
          (plotter): Plotter(count=0)
        )

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two outputs are equal or not.

        Args:
            other: The other output to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two outputs are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> output1 = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output2 = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output3 = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 0, 0]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output1.equal(output2)
        True
        >>> output1.equal(output3)
        False

        ```
        """

    @abstractmethod
    def get_content_generator(self, lazy: bool = True) -> BaseContentGenerator:
        r"""Get the HTML content generator associated to the output.

        Args:
            lazy: If ``True``, it forces the computation of the
                content, otherwise it returns a content generator
                object that contains the logic to generate the content.

        Returns:
            The HTML content generator.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> output = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output.get_content_generator()
        AccuracyContentGenerator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )

        ```
        """

    @abstractmethod
    def get_evaluator(self, lazy: bool = True) -> BaseEvaluator:
        r"""Get the evaluator associated to the output.

        Args:
            lazy: If ``True``, it forces the computation of the
                metrics, otherwise it returns an evaluator object
                that contains the logic to evaluate the metrics.

        Returns:
            The evaluator.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> output = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output.get_evaluator()
        AccuracyEvaluator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )

        ```
        """

    @abstractmethod
    def get_plotter(self, lazy: bool = True) -> BasePlotter:
        r"""Get the plotter associated to the output.

        Args:
            lazy: If ``True``, it forces the computation of the
                figures, otherwise it returns a plotter object
                that contains the logic to generate the figures.

        Returns:
            The plotter.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> output = AccuracyOutput(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> output.get_plotter()
        Plotter(count=0)

        ```
        """


class OutputEqualityComparator(BaseEqualityComparator[BaseOutput]):
    r"""Implement an equality comparator for ``BaseOutput`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> OutputEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseOutput, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseOutput):  # pragma: no cover
    EqualityTester.add_comparator(BaseOutput, OutputEqualityComparator())



r"""Implement a function to compute the Top-k Accuracy classification
metrics."""

from __future__ import annotations

__all__ = ["binary_top_k_accuracy", "multiclass_top_k_accuracy", "top_k_accuracy"]

from typing import TYPE_CHECKING

from sklearn import metrics

from arkas.metric.utils import (
    contains_nan,
    preprocess_score_binary,
    preprocess_score_multiclass,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np


def top_k_accuracy(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    k: Sequence[int] = (2,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the Area Under the Top-k Accuracy classification metrics.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. The binary case
            expects scores with shape ``(n_samples,)`` while the
            multiclass case expects scores with shape
            ``(n_samples, n_classes)``.
        k: The numbers of most likely outcomes considered to find the
            correct label.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import top_k_accuracy
    >>> # binary
    >>> metrics = top_k_accuracy(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1]), k=[1, 2]
    ... )
    >>> metrics
    {'count': 5, 'top_1_accuracy': 1.0, 'top_2_accuracy': 1.0}
    >>> # multiclass
    >>> metrics = top_k_accuracy(
    ...     y_true=np.array([0, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [[0.5, 0.2, 0.2], [0.3, 0.4, 0.2], [0.2, 0.4, 0.3], [0.7, 0.2, 0.1]]
    ...     ),
    ...     k=[1, 2, 3],
    ... )
    >>> metrics
    {'count': 4, 'top_1_accuracy': 0.5, 'top_2_accuracy': 0.75, 'top_3_accuracy': 1.0}

    ```
    """
    if y_score.ndim == 1:
        return binary_top_k_accuracy(
            y_true=y_true.ravel(),
            y_score=y_score.ravel(),
            k=k,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
    return multiclass_top_k_accuracy(
        y_true=y_true,
        y_score=y_score,
        k=k,
        prefix=prefix,
        suffix=suffix,
        nan_policy=nan_policy,
    )


def binary_top_k_accuracy(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    k: Sequence[int] = (2,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the Area Under the Top-k Accuracy classification metrics
    for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples,)``.
        k: The numbers of most likely outcomes considered to find the
            correct label.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.
    """
    y_true, y_score = preprocess_score_binary(
        y_true=y_true, y_score=y_score, drop_nan=nan_policy == "omit"
    )
    return _top_k_accuracy(
        y_true=y_true,
        y_score=y_score,
        k=k,
        prefix=prefix,
        suffix=suffix,
        nan_policy=nan_policy,
    )


def multiclass_top_k_accuracy(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    k: Sequence[int] = (2,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the Area Under the Top-k Accuracy classification metrics
    for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        k: The numbers of most likely outcomes considered to find the
            correct label.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.
    """
    y_true, y_score = preprocess_score_multiclass(y_true, y_score, drop_nan=nan_policy == "omit")
    return _top_k_accuracy(
        y_true=y_true,
        y_score=y_score,
        k=k,
        prefix=prefix,
        suffix=suffix,
        nan_policy=nan_policy,
    )


def _top_k_accuracy(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    k: Sequence[int] = (2,),
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the Area Under the Top-k Accuracy classification metrics
    for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        k: The numbers of most likely outcomes considered to find the
            correct label.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.
    """
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_score_nan = contains_nan(arr=y_score, nan_policy=nan_policy, name="'y_score'")

    count = y_true.size
    out = {}
    for _k in k:
        top_k_accuracy = float("nan")
        if count > 0 and not y_true_nan and not y_score_nan:
            top_k_accuracy = float(
                metrics.top_k_accuracy_score(y_true=y_true, y_score=y_score, k=_k)
            )
        out[f"{prefix}top_{_k}_accuracy{suffix}"] = top_k_accuracy
    return {f"{prefix}count{suffix}": count} | out



r"""Contain the energy distance between two 1D distributions
evaluator."""

from __future__ import annotations

__all__ = ["EnergyDistanceEvaluator"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from arkas.evaluator.lazy import BaseLazyEvaluator
from arkas.metric.utils import check_nan_policy
from arkas.result import EnergyDistanceResult, Result
from arkas.utils.array import to_array

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class EnergyDistanceEvaluator(BaseLazyEvaluator[EnergyDistanceResult]):
    r"""Implement the energy distance between two 1D distributions
    evaluator.

    Args:
        u_values: The values observed in the (empirical) distribution.
        v_values: The values observed in the (empirical) distribution.
        drop_nulls: If ``True``, the rows with null values in
            ``x`` or ``y`` columns are dropped.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import EnergyDistanceEvaluator
    >>> evaluator = EnergyDistanceEvaluator(u_values="target", v_values="pred")
    >>> evaluator
    EnergyDistanceEvaluator(u_values='target', v_values='pred', drop_nulls=True, nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [1, 2, 3, 4, 5], "target": [1, 2, 3, 4, 5]})
    >>> result = evaluator.evaluate(data)
    >>> result
    EnergyDistanceResult(u_values=(5,), v_values=(5,), nan_policy='propagate')

    ```
    """

    def __init__(
        self,
        u_values: str,
        v_values: str,
        drop_nulls: bool = True,
        nan_policy: str = "propagate",
    ) -> None:
        super().__init__(drop_nulls=drop_nulls)
        self._u_values = u_values
        self._v_values = v_values

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "u_values": self._u_values,
                "v_values": self._v_values,
                "drop_nulls": self._drop_nulls,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> EnergyDistanceResult | Result:
        logger.info(
            f"Evaluating the energy distance | u_values={self._u_values!r} | "
            f"v_values={self._v_values!r} | drop_nulls={self._drop_nulls} | "
            f"nan_policy={self._nan_policy!r}"
        )
        return self._evaluate(data, lazy)

    def _compute_result(self, data: pl.DataFrame) -> EnergyDistanceResult:
        return EnergyDistanceResult(
            u_values=to_array(data[self._u_values]).ravel(),
            v_values=to_array(data[self._v_values]).ravel(),
            nan_policy=self._nan_policy,
        )

    def _get_columns(self) -> tuple[str, ...]:
        return (self._u_values, self._v_values)



r"""Contain a base class that partially implements the lazy computation
logic."""

from __future__ import annotations

__all__ = ["BaseLazyOutput"]

from abc import abstractmethod
from typing import TYPE_CHECKING

from arkas.output.base import BaseOutput

if TYPE_CHECKING:
    from arkas.content.base import BaseContentGenerator
    from arkas.evaluator2.base import BaseEvaluator
    from arkas.output.vanilla import Output
    from arkas.plotter.base import BasePlotter


class BaseLazyOutput(BaseOutput):
    r"""Define a base class that partially implements the lazy
    computation logic."""

    def compute(self) -> Output:
        from arkas.output.vanilla import Output

        return Output(
            content=self.get_content_generator().compute(),
            evaluator=self.get_evaluator().compute(),
            plotter=self.get_plotter().compute(),
        )

    def get_content_generator(self, lazy: bool = True) -> BaseContentGenerator:
        content = self._get_content_generator()
        if not lazy:
            content = content.compute()
        return content

    def get_evaluator(self, lazy: bool = True) -> BaseEvaluator:
        evaluator = self._get_evaluator()
        if not lazy:
            evaluator = evaluator.compute()
        return evaluator

    def get_plotter(self, lazy: bool = True) -> BasePlotter:
        plotter = self._get_plotter()
        if not lazy:
            plotter = plotter.compute()
        return plotter

    @abstractmethod
    def _get_content_generator(self) -> BaseContentGenerator:
        r"""Get the content generator associated to the output.

        Returns:
            The content generator.
        """

    @abstractmethod
    def _get_evaluator(self) -> BaseEvaluator:
        r"""Get the evaluator associated to the output.

        Returns:
            The evaluator.
        """

    @abstractmethod
    def _get_plotter(self) -> BasePlotter:
        r"""Get the plotter associated to the output.

        Returns:
            The plotter.
        """



r"""Contain the implementation of a simple evaluator."""

from __future__ import annotations

__all__ = ["Evaluator"]

from typing import Any

from coola import objects_are_equal

from arkas.evaluator2.base import BaseEvaluator


class Evaluator(BaseEvaluator):
    r"""Implement a simple evaluator.

    Args:
        metrics: The dictionary of metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.evaluator2 import Evaluator
    >>> evaluator = Evaluator({"accuracy": 1.0, "total": 42})
    >>> evaluator
    Evaluator(count=2)
    >>> evaluator.evaluate()
    {'accuracy': 1.0, 'total': 42}

    ```
    """

    def __init__(self, metrics: dict | None = None) -> None:
        self._metrics = metrics or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={len(self._metrics):,})"

    def compute(self) -> Evaluator:
        return Evaluator(metrics=self._metrics)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._metrics, other._metrics, equal_nan=equal_nan)

    def evaluate(self, prefix: str = "", suffix: str = "") -> dict:
        return {f"{prefix}{key}{suffix}": value for key, value in self._metrics.items()}



r"""Contain the entrypoint to run evaluations or analyses."""

from __future__ import annotations

__all__ = ["main", "main_cli"]

import logging
from typing import Any
from unittest.mock import Mock

from iden.utils.time import timeblock

from arkas.cli.utils import log_run_info
from arkas.runner import setup_runner
from arkas.utils.imports import (
    is_hya_available,
    is_hydra_available,
    is_omegaconf_available,
)

if is_hydra_available():
    import hydra
else:  # pragma: no cover
    hydra = Mock()

if is_hya_available():
    from hya import register_resolvers
else:  # pragma: no cover

    def register_resolvers() -> None:
        return None


if is_omegaconf_available():
    from omegaconf import DictConfig, OmegaConf
else:  # pragma: no cover
    DictConfig = Mock()


logger = logging.getLogger(__name__)


def main(config: dict[str, Any]) -> None:
    r"""Initialize a runner given its configuration and execute its
    logic.

    Args:
        config: The dictionary with the configuration of the runner.
            This dictionary has to have a key ``'runner'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.cli.run import main
    >>> main(
    ...     {
    ...         "runner": {
    ...             "_target_": "arkas.runner.EvaluationRunner",
    ...             "ingestor": {
    ...                 "_target_": "grizz.ingestor.Ingestor",
    ...                 "frame": pl.DataFrame(
    ...                     {
    ...                         "pred": [3, 2, 0, 1, 0],
    ...                         "target": [3, 2, 0, 1, 0],
    ...                     }
    ...                 ),
    ...             },
    ...             "transformer": {"_target_": "grizz.transformer.DropDuplicate"},
    ...             "evaluator": {
    ...                 "_target_": "arkas.evaluator.AccuracyEvaluator",
    ...                 "y_true": "target",
    ...                 "y_pred": "pred",
    ...             },
    ...             "saver": {"_target_": "iden.io.PickleSaver"},
    ...             "path": "/tmp/data/metrics.pkl",
    ...         }
    ...     }
    ... )

    ```
    """
    with timeblock("Total time of the run: {time}"):
        logger.info("Creating runner...")
        runner = setup_runner(config["runner"])
        logger.info(f"runner:\n{runner}")
        logger.info("Start to execute the logic of the runner")
        runner.run()
        logger.info("End of the run")


@hydra.main(config_path=None, version_base=None)
def main_cli(config: DictConfig) -> None:
    r"""Define the CLI entrypoint to run an experiment.

    Please check the Hydra documentation to learn how Hydra works:
    https://hydra.cc/

    Args:
        config: The dictionary with the configuration of the runner.
            This dictionary has to have a key ``'runner'``.
    """
    log_run_info(config)
    register_resolvers()
    main(OmegaConf.to_container(config, resolve=True))


if __name__ == "__main__":  # pragma: no cover
    main_cli()



r"""Implement the balanced accuracy output."""

from __future__ import annotations

__all__ = ["BalancedAccuracyOutput"]

from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping

from arkas.content.balanced_accuracy import BalancedAccuracyContentGenerator
from arkas.evaluator2.balanced_accuracy import BalancedAccuracyEvaluator
from arkas.metric.utils import check_nan_policy
from arkas.output.lazy import BaseLazyOutput
from arkas.plotter.vanilla import Plotter

if TYPE_CHECKING:
    from arkas.state.accuracy import AccuracyState


class BalancedAccuracyOutput(BaseLazyOutput):
    r"""Implement the balanced accuracy output.

    Args:
        state: The state containing the ground truth and predicted
            labels.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.output import BalancedAccuracyOutput
    >>> from arkas.state import AccuracyState
    >>> output = BalancedAccuracyOutput(
    ...     AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> output
    BalancedAccuracyOutput(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )
    >>> output.get_content_generator()
    BalancedAccuracyContentGenerator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )
    >>> output.get_evaluator()
    BalancedAccuracyEvaluator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )
    >>> output.get_plotter()
    Plotter(count=0)

    ```
    """

    def __init__(self, state: AccuracyState, nan_policy: str = "propagate") -> None:
        self._state = state
        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"state": self._state, "nan_policy": self._nan_policy}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._state.equal(other._state, equal_nan=equal_nan)
            and self._nan_policy == other._nan_policy
        )

    def _get_content_generator(self) -> BalancedAccuracyContentGenerator:
        return BalancedAccuracyContentGenerator(state=self._state, nan_policy=self._nan_policy)

    def _get_evaluator(self) -> BalancedAccuracyEvaluator:
        return BalancedAccuracyEvaluator(state=self._state, nan_policy=self._nan_policy)

    def _get_plotter(self) -> Plotter:
        return Plotter()



r"""Contain data outputs."""

from __future__ import annotations

__all__ = [
    "AccuracyOutput",
    "BalancedAccuracyOutput",
    "BaseLazyOutput",
    "BaseOutput",
    "ColumnCooccurrenceOutput",
    "ContentOutput",
    "DataFrameSummaryOutput",
    "EmptyOutput",
    "Output",
    "OutputDict",
]

from arkas.output.accuracy import AccuracyOutput
from arkas.output.balanced_accuracy import BalancedAccuracyOutput
from arkas.output.base import BaseOutput
from arkas.output.column_cooccurrence import ColumnCooccurrenceOutput
from arkas.output.content import ContentOutput
from arkas.output.empty import EmptyOutput
from arkas.output.frame_summary import DataFrameSummaryOutput
from arkas.output.lazy import BaseLazyOutput
from arkas.output.mapping import OutputDict
from arkas.output.vanilla import Output



r"""Contain utility functions to generate histograms."""

from __future__ import annotations

__all__ = ["adjust_nbins", "find_nbins"]

import math

import numpy as np


def adjust_nbins(nbins: int | None, array: np.ndarray) -> int | None:
    r"""Return the adjusted number of bins.

    Args:
        nbins: The initial number of bins.
        array: The array with the value to plot in the histogram.

    Returns:
        The adjusted number of bins.

    Example usage:

    ```pycon

    >>> from arkas.plot.utils.hist import adjust_nbins
    >>> nbins = adjust_nbins(nbins=100, array=np.array([1, 4, 5, 6]))
    >>> nbins
    6

    ```
    """
    if array.size == 0 or nbins is None:
        return nbins
    if np.issubdtype(array.dtype, np.integer):
        return min(nbins, (np.max(array) - np.min(array) + 1).item())
    return nbins


def find_nbins(bin_size: float, min: float, max: float) -> int:  # noqa: A002
    r"""Find the number of bins from the bin size and the range of
    values.

    Args:
        bin_size: The target bin size.
        min: The minimum value.
        max: The maximum value.

    Returns:
        The number of bins.

    Raises:
        RuntimeError: if the bin size is invalid.
        RuntimeError: if the max value is invalid.

    Example usage:

    ```pycon

    >>> from arkas.plot.utils.hist import find_nbins
    >>> nbins = find_nbins(bin_size=1, min=0, max=10)
    >>> nbins
    11

    ```
    """
    if bin_size <= 0:
        msg = f"Incorrect bin_size {bin_size}. bin_size must be greater than 0"
        raise RuntimeError(msg)
    if max < min:
        msg = f"Incorrect max {max}. max must be greater or equal to min: {min}"
        raise RuntimeError(msg)
    if min == max:
        return 1
    return math.ceil((max - min + 1) / bin_size)



r"""Contain the implementation for matplotlib figures."""

from __future__ import annotations

__all__ = ["MatplotlibFigure", "MatplotlibFigureConfig"]

import base64
import io
import sys
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.figure.base import BaseFigure, BaseFigureConfig

if TYPE_CHECKING:
    from matplotlib.colors import Normalize

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )


class MatplotlibFigure(BaseFigure):
    r"""Implement the matplotlib figure.

    Args:
        figure: The matplotlib figure.
        reactive: If ``True``, the generated is configured to be
            reactive to the screen size.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.figure import MatplotlibFigure
    >>> fig = MatplotlibFigure(plt.subplots()[0])
    >>> fig
    MatplotlibFigure(reactive=True)

    ```
    """

    def __init__(self, figure: plt.Figure, reactive: bool = True) -> None:
        self._figure = figure
        self._reactive = reactive

    def __repr__(self) -> str:
        args = repr_mapping_line({"reactive": self._reactive})
        return f"{self.__class__.__qualname__}({args})"

    @property
    def figure(self) -> plt.Figure:
        return self._figure

    def close(self) -> None:
        plt.close(self._figure)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.figure, other.figure, equal_nan=equal_nan)
            and self._reactive == other._reactive
        )

    def set_reactive(self, reactive: bool) -> Self:
        return self.__class__(figure=self.figure, reactive=reactive)

    def to_html(self) -> str:
        self._figure.tight_layout()
        img = io.BytesIO()
        self._figure.savefig(img, format="png", bbox_inches="tight")
        img.seek(0)
        data = base64.b64encode(img.getvalue()).decode("utf-8")
        style = 'style="width:100%; height:auto;" ' if self._reactive else ""
        return f'<img {style}src="data:image/png;charset=utf-8;base64, {data}">'


class MatplotlibFigureConfig(BaseFigureConfig):
    r"""Implement the matplotlib figure config.

    Args:
        color_norm: The color normalization.
        **kwargs: Additional keyword arguments to pass to
            ``matplotlib.pyplot.subplots``.

    Example usage:

    ```pycon

    >>> from arkas.figure import MatplotlibFigureConfig
    >>> config = MatplotlibFigureConfig()
    >>> config
    MatplotlibFigureConfig(color_norm=None)

    ```
    """

    def __init__(self, color_norm: Normalize | None = None, **kwargs: Any) -> None:
        self._color_norm = color_norm
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line({"color_norm": self._color_norm} | self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    @classmethod
    def backend(cls) -> str:
        return "matplotlib"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        # color_norm is excluded from the comparison as it is not straightforward
        # to compare to Normalize objects.
        return objects_are_equal(self.get_args(), other.get_args(), equal_nan=equal_nan)

    def get_args(self) -> dict:
        return self._kwargs

    def get_color_norm(self) -> Normalize | None:
        r"""Get the color normalization.

        Returns:
            The color normalization.

        Example usage:

        ```pycon

        >>> from matplotlib.colors import LogNorm
        >>> from arkas.figure import MatplotlibFigureConfig
        >>> config = MatplotlibFigureConfig(color_norm=LogNorm())
        >>> config.get_color_norm()
        <matplotlib.colors.LogNorm object at 0x...>

        ```
        """
        return self._color_norm



r"""Contain the implementation of a HTML content generator that returns
the pairwise column co-occurrence."""

from __future__ import annotations

__all__ = ["ColumnCooccurrenceContentGenerator", "create_table", "create_template"]

import logging
from typing import TYPE_CHECKING, Any

import numpy as np
from coola import objects_are_equal
from coola.utils import str_indent
from grizz.utils.cooccurrence import compute_pairwise_cooccurrence
from jinja2 import Template

from arkas.content.section import BaseSectionContentGenerator
from arkas.figure.utils import figure2html
from arkas.plotter import ColumnCooccurrencePlotter

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from arkas.figure.base import BaseFigureConfig

logger = logging.getLogger(__name__)


class ColumnCooccurrenceContentGenerator(BaseSectionContentGenerator):
    r"""Implement a content generator that returns pairwise column co-
    occurrence.

    Args:
        frame: The DataFrame to analyze.
        ignore_self: If ``True``, the diagonal of the co-occurrence
            matrix (a.k.a. self-co-occurrence) is set to 0.
        figure_config: The figure configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.content import ColumnCooccurrenceContentGenerator
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> content = ColumnCooccurrenceContentGenerator(frame)
    >>> content
    ColumnCooccurrenceContentGenerator(shape=(7, 3), ignore_self=False)

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        ignore_self: bool = False,
        figure_config: BaseFigureConfig | None = None,
    ) -> None:
        self._frame = frame
        self._ignore_self = bool(ignore_self)
        self._figure_config = figure_config

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(shape={self._frame.shape}, "
            f"ignore_self={self._ignore_self})"
        )

    @property
    def frame(self) -> pl.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def ignore_self(self) -> bool:
        return self._ignore_self

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self.ignore_self == other.ignore_self
            and objects_are_equal(self.frame, other.frame, equal_nan=equal_nan)
            and objects_are_equal(self._figure_config, other._figure_config, equal_nan=equal_nan)
        )

    def generate_content(self) -> str:
        logger.info("Generating the DataFrame summary content...")
        figures = ColumnCooccurrencePlotter(
            frame=self._frame, ignore_self=self._ignore_self, figure_config=self._figure_config
        ).plot()
        columns = list(self._frame.columns)
        return Template(create_template()).render(
            {
                "nrows": f"{self._frame.shape[0]:,}",
                "ncols": f"{self._frame.shape[1]:,}",
                "columns": ", ".join([f"{x!r}" for x in columns]),
                "figure": figure2html(figures["column_cooccurrence"], close_fig=True),
                "table": create_table_section(
                    matrix=compute_pairwise_cooccurrence(
                        frame=self._frame, ignore_self=self._ignore_self
                    ),
                    columns=columns,
                ),
            }
        )


def create_template() -> str:
    r"""Return the template of the content.

    Returns:
        The content template.

    Example usage:

    ```pycon

    >>> from arkas.content.frame_summary import create_template
    >>> template = create_template()

    ```
    """
    return """This section shows an analysis of the pairwise column co-occurrence.
<ul>
  <li> number of columns: {{ncols}} </li>
  <li> number of rows: {{nrows}}</li>
</ul>
{{figure}}
<details>
    <summary>[show columns]</summary>
    {{columns}}
</details>
<p style="margin-top: 1rem;">
{{table}}
"""


def create_table_section(matrix: np.ndarray, columns: Sequence[str], top: int = 50) -> str:
    r"""Return the HTML code of the table section.

    Args:
        matrix: The co-occurrence matrix.
        columns: The column names.
        top: The number of co-occurrence pairs to show in the table.

    Returns:
        The HTML code of the table section.

    Example usage:

    ```pycon

    >>> from arkas.content.column_cooccurrence import create_table_section
    >>> section = create_table_section(
    ...     matrix=np.array([[5, 7, 1], [0, 6, 3], [8, 2, 4]]), columns=["col1", "col2", "col3"]
    ... )

    ```
    """
    if matrix.shape[0] == 0:
        return "<span>&#9888;</span> No table is generated because the column is empty"

    return Template(
        """<details>
    <summary>[show top-{{top}} pairwise column co-occurrence]</summary><br>
    The following table shows the top-{{top}} pairwise column co-occurrences.
    The co-occurrence matrix is symmetric and only the co-occurrences in the lower triangular matrix are shown.
    <ul>
      <li> <b>rank</b>: is the rank of the co-occurrence </li>
      <li> <b>column 1</b>: represents the first column of the co-occurrence matrix </li>
      <li> <b>column 2</b>: represents the second column of the co-occurrence matrix </li>
      <li> <b>count</b>: is the number of co-occurrences </li>
      <li> <b>percentage</b>: is the percentage of co-occurrences w.r.t. the total number of co-occurrences </li>
    </ul>

    {{table}}
</details>
"""
    ).render({"top": top, "table": create_table(matrix=matrix, columns=columns, top=top)})


def create_table(matrix: np.ndarray, columns: Sequence[str], top: int = 50) -> str:
    r"""Return the HTML code of the table.

    Args:
        matrix: The co-occurrence matrix.
        columns: The column names.
        top: The number of co-occurrence pairs to show in the table.

    Returns:
        The HTML code of the table.

    Example usage:

    ```pycon

    >>> from arkas.content.column_cooccurrence import create_table
    >>> section = create_table(
    ...     matrix=np.array([[5, 7, 1], [0, 6, 3], [8, 2, 4]]), columns=["col1", "col2", "col3"]
    ... )

    ```
    """
    total = np.triu(matrix).sum().item()
    # Fill the lower triangular part with -1 before to sort the occurrences
    half_matrix = np.triu(matrix) - np.tril(np.ones_like(matrix), -1)
    rows, cols = np.unravel_index(np.argsort(half_matrix, axis=None), matrix.shape)
    n = matrix.shape[0]
    m = min(top, int(n * (n + 1) / 2))
    rows, cols = rows[: -m - 1 : -1], cols[: -m - 1 : -1]
    table_rows = []
    for i, (r, c) in enumerate(zip(rows, cols)):
        table_rows.append(
            create_table_row(
                rank=i + 1, col1=columns[r], col2=columns[c], count=matrix[r, c].item(), total=total
            )
        )
    table_rows = "\n".join(table_rows)
    return Template(
        """<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr><th>rank</th><th>column 1</th><th>column 2</th><th>count</th><th>percentage</th></tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": str_indent(table_rows, num_spaces=8)})


def create_table_row(rank: int, col1: str, col2: str, count: int, total: int) -> str:
    r"""Return the HTML code of a table row.

    Args:
        rank: The rank of the pair of columns.
        col1: The first column.
        col2:  The second column.
        count: The number of co-occurrences.
        total: The total number of co-occurrences.

    Returns:
        The table row.

    Example usage:

    ```pycon

    >>> from arkas.content.column_cooccurrence import create_table_row
    >>> row = create_table_row(rank=2, col1="cat", col2="meow", count=42, total=100)

    ```
    """
    pct = 100 * count / total if total > 0 else float("nan")
    return Template(
        "<tr><th>{{rank}}</th>"
        "<td>{{col1}}</td>"
        "<td>{{col2}}</td>"
        '<td style="text-align: right;">{{count}}</td>'
        '<td style="text-align: right;">{{pct}}</td>'
        "</tr>"
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "rank": rank,
            "col1": col1,
            "col2": col2,
            "count": f"{count:,}",
            "pct": f"{pct:.4f} %",
        }
    )



r"""Implement the pairwise column co-occurrence output."""

from __future__ import annotations

__all__ = ["ColumnCooccurrenceOutput"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from arkas.content.column_cooccurrence import ColumnCooccurrenceContentGenerator
from arkas.evaluator2.vanilla import Evaluator
from arkas.output.lazy import BaseLazyOutput
from arkas.plotter.column_cooccurrence import ColumnCooccurrencePlotter

if TYPE_CHECKING:
    import polars as pl

    from arkas.figure.base import BaseFigureConfig


class ColumnCooccurrenceOutput(BaseLazyOutput):
    r"""Implement the pairwise column co-occurrence output.

    Args:
        frame: The DataFrame to analyze.
        ignore_self: If ``True``, the diagonal of the co-occurrence
            matrix (a.k.a. self-co-occurrence) is set to 0.
        figure_config: The figure configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.output import ColumnCooccurrenceOutput
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> output = ColumnCooccurrenceOutput(frame)
    >>> output
    ColumnCooccurrenceOutput(shape=(7, 3), ignore_self=False)
    >>> output.get_content_generator()
    ColumnCooccurrenceContentGenerator(shape=(7, 3), ignore_self=False)
    >>> output.get_evaluator()
    Evaluator(count=0)
    >>> output.get_plotter()
    ColumnCooccurrencePlotter(shape=(7, 3), ignore_self=False)

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        ignore_self: bool = False,
        figure_config: BaseFigureConfig | None = None,
    ) -> None:
        self._frame = frame
        self._ignore_self = bool(ignore_self)
        self._figure_config = figure_config

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(shape={self._frame.shape}, "
            f"ignore_self={self._ignore_self})"
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._ignore_self == other._ignore_self
            and objects_are_equal(self._frame, other._frame, equal_nan=equal_nan)
            and objects_are_equal(self._figure_config, other._figure_config, equal_nan=equal_nan)
        )

    def _get_content_generator(self) -> ColumnCooccurrenceContentGenerator:
        return ColumnCooccurrenceContentGenerator(
            frame=self._frame, ignore_self=self._ignore_self, figure_config=self._figure_config
        )

    def _get_evaluator(self) -> Evaluator:
        return Evaluator()

    def _get_plotter(self) -> ColumnCooccurrencePlotter:
        return ColumnCooccurrencePlotter(
            frame=self._frame, ignore_self=self._ignore_self, figure_config=self._figure_config
        )



r"""Implement the regression error metrics."""

from __future__ import annotations

__all__ = ["regression_errors"]


from typing import TYPE_CHECKING

from arkas.metric.regression.abs_error import mean_absolute_error, median_absolute_error
from arkas.metric.regression.mse import mean_squared_error

if TYPE_CHECKING:
    import numpy as np


def regression_errors(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the regression error metrics.

    Args:
        y_true: The ground truth target values.
        y_pred: The predicted values.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import regression_errors
    >>> regression_errors(y_true=np.array([1, 2, 3, 4, 5]), y_pred=np.array([1, 2, 3, 4, 5]))
    {'count': 5,
     'mean_absolute_error': 0.0,
     'median_absolute_error': 0.0,
     'mean_squared_error': 0.0}

    ```
    """
    return (
        mean_absolute_error(
            y_true=y_true,
            y_pred=y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
        | median_absolute_error(
            y_true=y_true,
            y_pred=y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
        | mean_squared_error(
            y_true=y_true,
            y_pred=y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=nan_policy,
        )
    )



r"""Contain an evaluator that evaluates a mapping of evaluators."""

from __future__ import annotations

__all__ = ["EvaluatorDict"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from arkas.evaluator import BaseEvaluator
from arkas.result import EmptyResult, MappingResult, Result

if TYPE_CHECKING:
    from collections.abc import Hashable, Mapping

    import polars as pl

    from arkas.result import BaseResult

logger = logging.getLogger(__name__)


class EvaluatorDict(BaseEvaluator):
    r"""Implement an evaluator that sequentially evaluates a mapping of
    evaluators.

    Args:
        evaluators: The mapping of evaluators to evaluate.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import (
    ...     EvaluatorDict,
    ...     BinaryPrecisionEvaluator,
    ...     BinaryRecallEvaluator,
    ... )
    >>> evaluator = EvaluatorDict(
    ...     {
    ...         "precision": BinaryPrecisionEvaluator(y_true="target", y_pred="pred"),
    ...         "recall": BinaryRecallEvaluator(y_true="target", y_pred="pred"),
    ...     }
    ... )
    >>> evaluator
    EvaluatorDict(
      (precision): BinaryPrecisionEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
      (recall): BinaryRecallEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
    )
    >>> data = pl.DataFrame({"pred": [1, 0, 0, 1, 1], "target": [1, 0, 0, 1, 1]})
    >>> result = evaluator.evaluate(data)
    >>> result
    MappingResult(count=2)
    >>> result = evaluator.evaluate(data, lazy=False)
    >>> result
    Result(metrics=2, figures=2)

    ```
    """

    def __init__(self, evaluators: Mapping[Hashable, BaseEvaluator]) -> None:
        self._evaluators = evaluators

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._evaluators))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> BaseResult:
        out = MappingResult(
            {
                key: evaluator.evaluate(data=data, lazy=lazy)
                for key, evaluator in self._evaluators.items()
            }
        )
        if lazy or isinstance(out, EmptyResult):
            return out
        return Result(metrics=out.compute_metrics(), figures=out.generate_figures())



r"""Implement the average precision result."""

from __future__ import annotations

__all__ = [
    "AveragePrecisionResult",
    "BaseAveragePrecisionResult",
    "BinaryAveragePrecisionResult",
    "MulticlassAveragePrecisionResult",
    "MultilabelAveragePrecisionResult",
]

from typing import TYPE_CHECKING, Any

import numpy as np
from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.classification.ap import (
    average_precision,
    binary_average_precision,
    find_label_type,
    multiclass_average_precision,
    multilabel_average_precision,
)
from arkas.metric.utils import (
    check_label_type,
    check_nan_policy,
    check_same_shape_score,
)
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import matplotlib.pyplot as plt


class AveragePrecisionResult(BaseResult):
    r"""Implement the average precision result.

    This result can be used in 3 different settings:

    - binary: ``y_true`` must be an array of shape ``(*)``
        with ``0`` and ``1`` values, and ``y_score`` must be an array
        of shape ``(*)``.
    - multiclass: ``y_true`` must be an array of shape ``(n_samples,)``
        with values in ``{0, ..., n_classes-1}``, and ``y_score`` must
        be an array of shape ``(n_samples, n_classes)``.
    - multilabel: ``y_true`` must be an array of shape
        ``(n_samples, n_classes)`` with ``0`` and ``1`` values, and
        ``y_score`` must be an array of shape
        ``(n_samples, n_classes)``.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        label_type: The type of labels used to evaluate the metrics.
            The valid values are: ``'binary'``, ``'multiclass'``,
            ``'multilabel'``, and ``'auto'``. If ``'binary'`` or
            ``'multilabel'``, ``y_true`` values  must be ``0`` and
            ``1``. If ``'auto'``, it tries to automatically find the
            label type from the arrays' shape.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import AveragePrecisionResult
    >>> # binary
    >>> result = AveragePrecisionResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_score=np.array([2, -1, 0, 3, 1]),
    ...     label_type="binary",
    ... )
    >>> result
    AveragePrecisionResult(y_true=(5,), y_score=(5,), label_type='binary', nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': 1.0, 'count': 5}
    >>> # multilabel
    >>> result = AveragePrecisionResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_score=np.array([[2, -1, -1], [-1, 1, 2], [0, 2, 3], [3, -2, -4], [1, -3, -5]]),
    ...     label_type="multilabel",
    ... )
    >>> result
    AveragePrecisionResult(y_true=(5, 3), y_score=(5, 3), label_type='multilabel', nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': array([1. , 1. , 0.477...]),
     'count': 5,
     'macro_average_precision': 0.825...,
     'micro_average_precision': 0.588...,
     'weighted_average_precision': 0.804...}
    >>> # multiclass
    >>> result = AveragePrecisionResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [
    ...             [0.7, 0.2, 0.1],
    ...             [0.4, 0.3, 0.3],
    ...             [0.1, 0.8, 0.1],
    ...             [0.2, 0.3, 0.5],
    ...             [0.4, 0.4, 0.2],
    ...             [0.1, 0.2, 0.7],
    ...         ]
    ...     ),
    ...     label_type="multiclass",
    ... )
    >>> result
    AveragePrecisionResult(y_true=(6,), y_score=(6, 3), label_type='multiclass', nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': array([0.833..., 0.75 , 0.75 ]),
     'count': 6,
     'macro_average_precision': 0.777...,
     'micro_average_precision': 0.75,
     'weighted_average_precision': 0.777...}
    >>> # auto
    >>> result = AveragePrecisionResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_score=np.array([2, -1, 0, 3, 1]),
    ... )
    >>> result
    AveragePrecisionResult(y_true=(5,), y_score=(5,), label_type='binary', nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': 1.0, 'count': 5}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        label_type: str = "auto",
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_score = y_score.astype(np.float64)
        self._label_type = (
            find_label_type(y_true=y_true, y_score=y_score) if label_type == "auto" else label_type
        )

        self._check_inputs()

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_score": self._y_score.shape,
                "label_type": self._label_type,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def label_type(self) -> str:
        return self._label_type

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_score(self) -> np.ndarray:
        return self._y_score

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return average_precision(
            y_true=self._y_true,
            y_score=self._y_score,
            label_type=self._label_type,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_score, other.y_score, equal_nan=equal_nan)
            and self.label_type == other.label_type
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}

    def _check_inputs(self) -> None:
        if self._y_true.ndim not in {1, 2}:
            msg = (
                f"'y_true' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_true.shape}"
            )
            raise ValueError(msg)
        if self._y_score.ndim not in {1, 2}:
            msg = (
                f"'y_score' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_score.shape}"
            )
            raise ValueError(msg)
        if self._y_true.ndim == self._y_score.ndim and self._y_true.shape != self._y_score.shape:
            msg = (
                f"'y_true' and 'y_score' have different shapes: {self._y_true.shape} vs "
                f"{self._y_score.shape}"
            )
            raise ValueError(msg)
        check_label_type(self._label_type)


class BaseAveragePrecisionResult(BaseResult):
    r"""Implement the base class to implement the average precision
    results.

    Args:
        y_true: The ground truth target labels.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryAveragePrecisionResult
    >>> result = BinaryAveragePrecisionResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryAveragePrecisionResult(y_true=(5,), y_score=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': 1.0, 'count': 5}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_score = y_score

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_score": self._y_score.shape,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_score(self) -> np.ndarray:
        return self._y_score

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_score, other.y_score, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )


class BinaryAveragePrecisionResult(BaseAveragePrecisionResult):
    r"""Implement the average precision result for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with ``0`` and
            ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, *)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryAveragePrecisionResult
    >>> result = BinaryAveragePrecisionResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1])
    ... )
    >>> result
    BinaryAveragePrecisionResult(y_true=(5,), y_score=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': 1.0, 'count': 5}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_score(y_true, y_score)
        super().__init__(y_true=y_true.ravel(), y_score=y_score.ravel(), nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return binary_average_precision(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}


class MulticlassAveragePrecisionResult(BaseAveragePrecisionResult):
    r"""Implement the average precision result for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` with ``0`` and
            ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MulticlassAveragePrecisionResult
    >>> result = MulticlassAveragePrecisionResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [
    ...             [0.7, 0.2, 0.1],
    ...             [0.4, 0.3, 0.3],
    ...             [0.1, 0.8, 0.1],
    ...             [0.2, 0.5, 0.3],
    ...             [0.3, 0.3, 0.4],
    ...             [0.1, 0.2, 0.7],
    ...         ]
    ...     ),
    ... )
    >>> result
    MulticlassAveragePrecisionResult(y_true=(6,), y_score=(6, 3), nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': array([1., 1., 1.]),
     'count': 6,
     'macro_average_precision': 1.0,
     'micro_average_precision': 1.0,
     'weighted_average_precision': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        y_true = y_true.ravel()
        if y_true.shape[0] != y_score.shape[0]:
            msg = (
                f"'y_true' and 'y_score' have different first dimension: {y_true.shape} vs "
                f"{y_score.shape}"
            )
            raise RuntimeError(msg)
        super().__init__(y_true=y_true, y_score=y_score, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multiclass_average_precision(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}


class MultilabelAveragePrecisionResult(BaseAveragePrecisionResult):
    r"""Implement the average precision result for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)`` with ``0``
            and ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MultilabelAveragePrecisionResult
    >>> result = MultilabelAveragePrecisionResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_score=np.array([[2, -1, 1], [-1, 1, -2], [0, 2, -3], [3, -2, 4], [1, -3, 5]]),
    ... )
    >>> result
    MultilabelAveragePrecisionResult(y_true=(5, 3), y_score=(5, 3), nan_policy='propagate')
    >>> result.compute_metrics()
    {'average_precision': array([1., 1., 1.]),
     'count': 5,
     'macro_average_precision': 1.0,
     'micro_average_precision': 1.0,
     'weighted_average_precision': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_score(y_true, y_score)
        super().__init__(y_true=y_true, y_score=y_score, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multilabel_average_precision(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}



r"""Contain the base class to implement an output exporter."""

from __future__ import annotations

__all__ = ["BaseExporter", "is_exporter_config", "setup_exporter"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from arkas.output.base import BaseOutput

logger = logging.getLogger(__name__)


class BaseExporter(ABC, metaclass=AbstractFactory):
    r"""Define the base class to export an output object.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> import numpy as np
    >>> from arkas.output import AccuracyOutput
    >>> from arkas.state import AccuracyState
    >>> from arkas.exporter import MetricExporter
    >>> output = AccuracyOutput(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     exporter = MetricExporter(path=Path(tmpdir).joinpath("metrics.pkl"))
    ...     print(exporter)
    ...     exporter.export(output)
    ...
    MetricExporter(
      (path): .../metrics.pkl
      (saver): PickleSaver(protocol=5)
      (exist_ok): False
      (show_metrics): True
    )

    ```
    """

    @abstractmethod
    def export(self, output: BaseOutput) -> None:
        r"""Export an output.

        Args:
            output: The output object to export.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> import numpy as np
        >>> from arkas.output import AccuracyOutput
        >>> from arkas.state import AccuracyState
        >>> from arkas.exporter import MetricExporter
        >>> output = AccuracyOutput(
        ...     state=AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     exporter = MetricExporter(path=Path(tmpdir).joinpath("metrics.pkl"))
        ...     exporter.export(output)
        ...

        ```
        """


def is_exporter_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseExporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseExporter`` object.

    Example usage:

    ```pycon

    >>> from arkas.exporter import is_exporter_config
    >>> is_exporter_config(
    ...     {
    ...         "_target_": "arkas.exporter.MetricExporter",
    ...         "path": "/path/to/data.csv",
    ...     }
    ... )
    True

    ```
    """
    return is_object_config(config, BaseExporter)


def setup_exporter(
    exporter: BaseExporter | dict,
) -> BaseExporter:
    r"""Set up a exporter.

    The exporter is instantiated from its configuration
    by using the ``BaseExporter`` factory function.

    Args:
        exporter: A exporter or its configuration.

    Returns:
        An instantiated exporter.

    Example usage:

    ```pycon

    >>> from arkas.exporter import setup_exporter
    >>> exporter = setup_exporter(
    ...     {
    ...         "_target_": "arkas.exporter.MetricExporter",
    ...         "path": "/path/to/data.csv",
    ...     }
    ... )
    >>> exporter
    MetricExporter(
      (path): /path/to/data.csv
      (saver): PickleSaver(protocol=5)
      (exist_ok): False
      (show_metrics): True
    )

    ```
    """
    if isinstance(exporter, dict):
        logger.info("Initializing a exporter from its configuration... ")
        exporter = BaseExporter.factory(**exporter)
    if not isinstance(exporter, BaseExporter):
        logger.warning(f"exporter is not a `BaseExporter` (received: {type(exporter)})")
    return exporter



r"""Contain the base class to implement a HTML Content Generator."""

from __future__ import annotations

__all__ = ["BaseContentGenerator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from collections.abc import Sequence

    from coola.equality import EqualityConfig


class BaseContentGenerator(ABC):
    r"""Define the base class to implement a HTML Content Generator.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.content import AccuracyContentGenerator
    >>> from arkas.state import AccuracyState
    >>> generator = AccuracyContentGenerator(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> generator
    AccuracyContentGenerator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    @abstractmethod
    def compute(self) -> BaseContentGenerator:
        r"""Compute the content and return a new content generator.

        Returns:
            A new content generator with the computed content.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.content import AccuracyContentGenerator
        >>> from arkas.state import AccuracyState
        >>> generator = AccuracyContentGenerator(
        ...     state=AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator
        AccuracyContentGenerator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )
        >>> generator2 = generator.compute()
        >>> generator2
        ContentGenerator()

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two content generators are equal or not.

        Args:
            other: The other content generator to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two content generators are equal,
                otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.content import AccuracyContentGenerator
        >>> from arkas.state import AccuracyState
        >>> generator1 = AccuracyContentGenerator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator2 = AccuracyContentGenerator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator3 = AccuracyContentGenerator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 0, 0]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator1.equal(generator2)
        True
        >>> generator1.equal(generator3)
        False

        ```
        """

    @abstractmethod
    def generate_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        r"""Return the HTML body associated to the content.

        Args:
            number: The section number, if any.
            tags: The tags associated to the content section, if any.
            depth: The depth in the content section, if any.

        Returns:
            The HTML body associated to the content section.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.content import AccuracyContentGenerator
        >>> from arkas.state import AccuracyState
        >>> generator = AccuracyContentGenerator(
        ...     state=AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator.generate_body()

        ```
        """

    @abstractmethod
    def generate_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        r"""Return the HTML table of content (TOC) associated to the
        section.

        Args:
            number: The section number associated to the
                section, if any.
            tags: The tags associated to the section, if any.
            depth: The depth in the report, if any.
            max_depth: The maximum depth to generate in the TOC.

        Returns:
            The HTML table of content associated to the section.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.content import AccuracyContentGenerator
        >>> from arkas.state import AccuracyState
        >>> generator = AccuracyContentGenerator(
        ...     state=AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> generator.generate_toc()

        ```
        """


class ContentGeneratorEqualityComparator(BaseEqualityComparator[BaseContentGenerator]):
    r"""Implement an equality comparator for ``BaseContentGenerator``
    objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> ContentGeneratorEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseContentGenerator, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseContentGenerator):  # pragma: no cover
    EqualityTester.add_comparator(BaseContentGenerator, ContentGeneratorEqualityComparator())



r"""Contain the implementation of a section that shows the results in a
HTML format."""

from __future__ import annotations

__all__ = ["ResultSection", "create_section_template"]

import logging
from typing import TYPE_CHECKING, Any

from coola.nested import to_flat_dict
from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping
from jinja2 import Template

from arkas.section.base import BaseSection
from arkas.utils.figure import figure2html
from arkas.utils.html import GO_TO_TOP, render_toc, tags2id, tags2title, valid_h_tag

if TYPE_CHECKING:
    from collections.abc import Sequence

    from arkas.result import BaseResult


logger = logging.getLogger(__name__)


class ResultSection(BaseSection):
    r"""Implement a section that show the results in a HTML format.

    Args:
        result: The data structure containing the results.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.section import ResultSection
    >>> from arkas.result import AccuracyResult
    >>> section = ResultSection(
    ...     result=AccuracyResult(
    ...         y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ...     )
    ... )
    >>> section
    ResultSection(
      (result): AccuracyResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    )
    >>> section.generate_html_body()

    ```
    """

    def __init__(self, result: BaseResult) -> None:
        self._result = result

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"result": self._result}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(str_mapping({"result": self._result}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._result.equal(other._result)

    def generate_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Generating the accuracy section...")
        metrics = self._result.compute_metrics()
        figures = self._result.generate_figures()
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "metrics": create_table_metrics(metrics),
                "figures": create_figures(figures),
            }
        )

    def generate_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from arkas.section.accuracy import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">

{{metrics}}

{{figures}}

<p style="margin-top: 1rem;">
"""


def create_table_metrics(metrics: dict) -> str:
    r"""Return the HTML code of a table with the metrics.

    Args:
        metrics: The metrics.

    Returns:
        The HTML code of the table.

    Example usage:

    ```pycon

    >>> from arkas.section.result import create_table_metrics
    >>> row = create_table_metrics({"accuracy": 0.42, "f1": 0.57})

    ```
    """
    metrics = to_flat_dict(metrics)
    rows = [create_table_metrics_row(name=name, value=value) for name, value in metrics.items()]
    return Template(
        """<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>name</th>
            <th>value</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": "\n".join(rows)})


def create_table_metrics_row(name: str, value: Any) -> str:
    r"""Return the HTML code of a table row.

    Args:
        name: The metric name.
        value: The metric value.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> from arkas.section.result import create_table_metrics_row
    >>> row = create_table_metrics_row(name="accuracy", value=42.0)

    ```
    """
    return f'<tr><th>{name}</th><td style="text-align: right;">{value}</td></tr>'


def create_figures(figures: dict) -> str:
    r"""Return the HTML code of the figures.

    Args:
        figures: The figures.

    Returns:
        The HTML code of the figures.

    Example usage:

    ```pycon

    >>> from arkas.section.result import create_figures
    >>> from matplotlib import pyplot as plt
    >>> fig, _ = plt.subplots()
    >>> out = create_figures({"pr": fig})

    ```
    """
    figures = to_flat_dict(figures)
    rows = "\n".join(
        [
            f"<li>{name}</li>\n{figure2html(value, reactive=False, close_fig=True)}"
            for name, value in figures.items()
        ]
    )
    return f"<ul>\n{rows}\n</ul>"



r"""Contain the Kullback-Leibler (KL) divergence between two
distributions evaluator."""

from __future__ import annotations

__all__ = ["KLDivEvaluator"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from arkas.evaluator.lazy import BaseLazyEvaluator
from arkas.result import KLDivResult, Result
from arkas.utils.array import to_array

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class KLDivEvaluator(BaseLazyEvaluator[KLDivResult]):
    r"""Implement the Kullback-Leibler (KL) divergence between two
    distributions evaluator.

    Args:
        p: The values observed in the (empirical) distribution.
        q: The values observed in the (empirical) distribution.
        drop_nulls: If ``True``, the rows with null values in
            ``x`` or ``y`` columns are dropped.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import KLDivEvaluator
    >>> evaluator = KLDivEvaluator(p="target", q="pred")
    >>> evaluator
    KLDivEvaluator(p='target', q='pred', drop_nulls=True)
    >>> data = pl.DataFrame({"pred": [1, 2, 3, 4, 5], "target": [1, 2, 3, 4, 5]})
    >>> result = evaluator.evaluate(data)
    >>> result
    KLDivResult(p=(5,), q=(5,))

    ```
    """

    def __init__(self, p: str, q: str, drop_nulls: bool = True) -> None:
        super().__init__(drop_nulls=drop_nulls)
        self._p = p
        self._q = q

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "p": self._p,
                "q": self._q,
                "drop_nulls": self._drop_nulls,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> KLDivResult | Result:
        logger.info(
            f"Evaluating the Wasserstein distance | p={self._p!r} | "
            f"q={self._q!r} | drop_nulls={self._drop_nulls}"
        )
        return self._evaluate(data, lazy)

    def _compute_result(self, data: pl.DataFrame) -> KLDivResult:
        return KLDivResult(
            p=to_array(data[self._p]).ravel(),
            q=to_array(data[self._q]).ravel(),
        )

    def _get_columns(self) -> tuple[str, ...]:
        return (self._p, self._q)



r"""Contain the implementation of a simple HTML content generator."""

from __future__ import annotations

__all__ = ["ContentGenerator"]

import logging
from typing import Any

from arkas.content.section import BaseSectionContentGenerator

logger = logging.getLogger(__name__)


class ContentGenerator(BaseSectionContentGenerator):
    r"""Implement a section that analyze accuracy states.

    Args:
        content: The HTML content.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.content import ContentGenerator
    >>> generator = ContentGenerator("meow")
    >>> generator
    ContentGenerator()
    >>> generator.generate_content()
    'meow'

    ```
    """

    def __init__(self, content: str = "") -> None:
        self._content = str(content)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:  # noqa: ARG002
        if not isinstance(other, self.__class__):
            return False
        return self._content == other._content

    def generate_content(self) -> str:
        return self._content



r"""Contain utility functions to compute metrics."""

from __future__ import annotations

__all__ = [
    "check_array_ndim",
    "check_label_type",
    "check_nan_policy",
    "check_nan_pred",
    "check_same_shape",
    "check_same_shape_pred",
    "check_same_shape_score",
    "contains_nan",
    "multi_isnan",
    "preprocess_pred",
    "preprocess_pred_multilabel",
    "preprocess_same_shape_arrays",
    "preprocess_score_binary",
    "preprocess_score_multiclass",
    "preprocess_score_multilabel",
]

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence


def check_array_ndim(arr: np.ndarray, ndim: int) -> None:
    r"""Check if the number of array dimensions is matching the target
    number of dimensions.

    Args:
        arr: The array to check.
        ndim: The targeted number of array dimensions.

    Raises:
        ValueError: if the number of array dimensions does not match.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import check_array_ndim
    >>> check_array_ndim(np.ones((2, 3)), ndim=2)

    ```
    """
    if arr.ndim != ndim:
        msg = f"Incorrect number of array dimensions: {arr.ndim} (expected: {ndim})"
        raise ValueError(msg)


def check_label_type(label_type: str) -> None:
    r"""Check if the label type value is valid or not.

    Args:
        label_type: The type of labels.
            The valid values are ``'binary'``, ``'multiclass'``,
            ``'multilabel'``, and ``'auto'``.

    Raises:
        ValueError: if an invalid value is passed to ``label_type``.

    Example usage:

    ```pycon

    >>> from arkas.metric.utils import check_label_type
    >>> check_label_type(label_type="binary")

    ```
    """
    if label_type not in {"binary", "multiclass", "multilabel", "auto"}:
        msg = (
            f"Incorrect 'label_type': {label_type}. The supported label types are: "
            f"'binary', 'multiclass', 'multilabel', and 'auto'"
        )
        raise ValueError(msg)


def check_nan_policy(nan_policy: str) -> None:
    r"""Check the NaN policy.

    Args:
        nan_policy: The NaN policy.

    Raises:
        ValueError: if ``nan_policy`` is not ``'omit'``,
            ``'propagate'``, or ``'raise'``.

    Example usage:

    ```pycon

    >>> from arkas.metric.utils import check_nan_policy
    >>> check_nan_policy(nan_policy="omit")

    ```
    """
    if nan_policy not in {"omit", "propagate", "raise"}:
        msg = (
            f"Incorrect 'nan_policy': {nan_policy}. The valid values are: "
            f"'omit', 'propagate', 'raise'"
        )
        raise ValueError(msg)


def check_nan_pred(y_true: np.ndarray, y_pred: np.ndarray) -> None:
    r"""Check if any array elements in ``y_true`` or ``y_pred`` arrays is
    a NaN value.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.

    Raises:
        RuntimeError: ``'y_true'`` or ``'y_pred'`` has a NaN value.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import check_nan_pred
    >>> y_true = np.array([1, 0, 0, 1])
    >>> y_pred = np.array([0, 1, 0, 1])
    >>> check_nan_pred(y_true, y_pred)

    ```
    """
    if np.isnan(y_true).any():
        msg = "'y_true' contains at least one NaN value"
        raise RuntimeError(msg)
    if np.isnan(y_pred).any():
        msg = "'y_pred' contains at least one NaN value"
        raise RuntimeError(msg)


def check_same_shape(arrays: Iterable[np.ndarray]) -> None:
    r"""Check if arrays have the same shape.

    Args:
        arrays: The arrays to check.

    Raises:
        RuntimeError: if the arrays have different shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import check_same_shape
    >>> check_same_shape([np.array([1, 0, 0, 1]), np.array([0, 1, 0, 1])])

    ```
    """
    shapes = {arr.shape for arr in arrays}
    if len(shapes) > 1:
        msg = f"arrays have different shapes: {shapes}"
        raise RuntimeError(msg)


def check_same_shape_pred(y_true: np.ndarray, y_pred: np.ndarray) -> None:
    r"""Check if ``y_true`` and ``y_pred`` arrays have the same shape.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.

    Raises:
        RuntimeError: ``'y_true'`` and ``'y_pred'`` have different
            shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import check_same_shape_pred
    >>> y_true = np.array([1, 0, 0, 1])
    >>> y_pred = np.array([0, 1, 0, 1])
    >>> check_same_shape_pred(y_true, y_pred)

    ```
    """
    if y_true.shape != y_pred.shape:
        msg = f"'y_true' and 'y_pred' have different shapes: {y_true.shape} vs {y_pred.shape}"
        raise RuntimeError(msg)


def check_same_shape_score(y_true: np.ndarray, y_score: np.ndarray) -> None:
    r"""Check if ``y_true`` and ``y_score`` arrays have the same shape.

    Args:
        y_true: The ground truth target labels.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions.

    Raises:
        RuntimeError: ``'y_true'`` and ``'y_score'`` have different
            shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import check_same_shape_score
    >>> y_true = np.array([1, 0, 0, 1])
    >>> y_score = np.array([0, 1, 0, 1])
    >>> check_same_shape_score(y_true, y_score)

    ```
    """
    if y_true.shape != y_score.shape:
        msg = f"'y_true' and 'y_score' have different shapes: {y_true.shape} vs {y_score.shape}"
        raise RuntimeError(msg)


def contains_nan(arr: np.ndarray, nan_policy: str = "propagate", name: str = "input array") -> bool:
    r"""Indicate if the given array contains at least one NaN value.

    Args:
        arr: The array to check.
        nan_policy: The NaN policy. The valid values are ``'omit'``,
            ``'propagate'``, or ``'raise'``.
        name: An optional name to be more precise about the array when
            the exception is raised.

    Returns:
        ``True`` if the array contains at least one NaN value.

    Raises:
        ValueError: if the array contains at least one NaN value and
            ``nan_policy`` is ``'raise'``.
    """
    check_nan_policy(nan_policy)
    isnan = np.any(np.isnan(arr))
    if isnan and nan_policy == "raise":
        msg = f"{name} contains at least one NaN value"
        raise ValueError(msg)
    return isnan


def multi_isnan(arrays: Sequence[np.ndarray]) -> np.ndarray:
    r"""Test element-wise for NaN for all input arrays and return result
    as a boolean array.

    Args:
        arrays: The input arrays to test. All the arrays must have the
            same shape.

    Returns:
        A boolean array. ``True`` where any array is NaN,
            ``False`` otherwise.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import multi_isnan
    >>> mask = multi_isnan(
    ...     [np.array([1, 0, 0, 1, float("nan")]), np.array([1, float("nan"), 0, 1, 1])]
    ... )
    >>> mask
    array([False,  True, False, False,  True])

    ```
    """
    if len(arrays) == 0:
        msg = "'arrays' cannot be empty"
        raise RuntimeError(msg)
    mask = np.isnan(arrays[0])
    for arr in arrays[1:]:
        mask = np.logical_or(mask, np.isnan(arr))
    return mask


def preprocess_pred(
    y_true: np.ndarray, y_pred: np.ndarray, drop_nan: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    r"""Preprocess ``y_true`` and ``y_pred`` arrays.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_pred``
            arrays.

    Raises:
        RuntimeError: if an invalid value is passed to ``nan``.
        RuntimeError: ``'y_true'`` and ``'y_pred'`` have different
            shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_pred
    >>> y_true = np.array([1, 0, 0, 1, 1, float("nan")])
    >>> y_pred = np.array([0, 1, 0, 1, float("nan"), 1])
    >>> preprocess_pred(y_true, y_pred)
    (array([ 1.,  0.,  0.,  1.,  1., nan]), array([ 0.,  1.,  0.,  1., nan,  1.]))
    >>> preprocess_pred(y_true, y_pred, drop_nan=True)
    (array([1., 0., 0., 1.]), array([0., 1., 0., 1.]))

    ```
    """
    check_same_shape_pred(y_true, y_pred)
    if not drop_nan:
        return y_true, y_pred
    mask = np.logical_not(multi_isnan([y_true, y_pred]))
    return y_true[mask], y_pred[mask]


def preprocess_pred_multilabel(
    y_true: np.ndarray, y_pred: np.ndarray, drop_nan: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    r"""Preprocess ``y_true`` and ``y_pred`` arrays.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_pred``
            arrays.

    Raises:
        RuntimeError: if an invalid value is passed to ``nan``.
        RuntimeError: ``'y_true'`` and ``'y_pred'`` have different
            shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_pred_multilabel
    >>> y_true = np.array([[1, float("nan"), 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]])
    >>> y_pred = np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, float("nan")]])
    >>> preprocess_pred_multilabel(y_true, y_pred)
    (array([[ 1., nan,  1.],
            [ 0.,  1.,  0.],
            [ 0.,  1.,  0.],
            [ 1.,  0.,  1.],
            [ 1.,  0.,  1.]]),
     array([[ 1.,  0.,  1.],
            [ 0.,  1.,  0.],
            [ 0.,  1.,  0.],
            [ 1.,  0.,  1.],
            [ 1.,  0., nan]]))
    >>> preprocess_pred_multilabel(y_true, y_pred, drop_nan=True)
    (array([[0., 1., 0.],
            [0., 1., 0.],
            [1., 0., 1.]]),
     array([[0., 1., 0.],
            [0., 1., 0.],
            [1., 0., 1.]]))

    ```
    """
    if y_true.size == 0 and y_pred.size == 0:
        return np.array([]), np.array([])

    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))
    if y_true.ndim != 2:
        msg = f"'y_true' must be a 1d or 2d array but received an array of shape: {y_true.shape}"
        raise RuntimeError(msg)
    if y_pred.ndim == 1:
        y_pred = y_pred.reshape((-1, 1))
    check_same_shape_pred(y_true, y_pred)
    if not drop_nan:
        return y_true, y_pred

    mask = np.logical_not(np.logical_or(np.isnan(y_true).any(axis=1), np.isnan(y_pred).any(axis=1)))
    return y_true[mask], y_pred[mask]


def preprocess_same_shape_arrays(
    arrays: Sequence[np.ndarray], drop_nan: bool = False
) -> tuple[np.ndarray, ...]:
    r"""Preprocess a sequence of same shape arrays.

    Args:
        arrays: The arrays to preprocess.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_pred``
            arrays.

    Raises:
        RuntimeError: if the arrays have different shapes.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_same_shape_arrays
    >>> arrays = [
    ...     np.array([1, 0, 0, 1, 1, float("nan")]),
    ...     np.array([0, 1, 0, 1, float("nan"), 1]),
    ... ]
    >>> preprocess_same_shape_arrays(arrays)
    (array([ 1.,  0.,  0.,  1.,  1., nan]), array([ 0.,  1.,  0.,  1., nan,  1.]))
    >>> preprocess_same_shape_arrays(arrays, drop_nan=True)
    (array([1., 0., 0., 1.]), array([0., 1., 0., 1.]))

    ```
    """
    check_same_shape(arrays)
    if not drop_nan:
        return tuple(arrays)
    mask = np.logical_not(multi_isnan(arrays))
    return tuple(arr[mask] for arr in arrays)


def preprocess_score_binary(
    y_true: np.ndarray, y_score: np.ndarray, drop_nan: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    r"""Preprocess ``y_true`` and ``y_score`` arrays for the binary
    classification case.

    Args:
        y_true: The ground truth target labels. This input must be an
            array of shape ``(n_samples, *)``.
        y_score: The predicted labels. This input must be an
            array of shape ``(n_samples, *)``.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_score``
            arrays.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_score_binary
    >>> y_true = np.array([1, 0, 0, 1, 1, float("nan")])
    >>> y_score = np.array([0, 1, 0, 1, float("nan"), 1])
    >>> preprocess_score_binary(y_true, y_score)
    (array([ 1.,  0.,  0.,  1.,  1., nan]), array([ 0.,  1.,  0.,  1., nan,  1.]))
    >>> preprocess_score_binary(y_true, y_score, drop_nan=True)
    (array([1., 0., 0., 1.]), array([0., 1., 0., 1.]))

    ```
    """
    y_true, y_score = y_true.ravel(), y_score.ravel()
    if y_true.size == 0 and y_score.size == 0:
        return y_true, y_score

    check_same_shape_score(y_true, y_score)
    if not drop_nan:
        return y_true, y_score

    # Remove NaN values
    mask = np.logical_not(multi_isnan([y_true, y_score]))
    return y_true[mask], y_score[mask]


def preprocess_score_multiclass(
    y_true: np.ndarray, y_score: np.ndarray, drop_nan: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    r"""Preprocess ``y_true`` and ``y_score`` arrays for the multiclass
    classification case.

    Args:
        y_true: The ground truth target labels. This input must be an
            array of shape ``(n_samples,)`` or ``(n_samples, 1)``.
        y_score: The predicted labels. This input must be an
            array of shape ``(n_samples, n_classes)``.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_score``
            arrays.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_score_multiclass
    >>> y_true = np.array([0, 0, 1, 1, 2, float("nan")])
    >>> y_score = np.array(
    ...     [
    ...         [0.7, 0.2, 0.1],
    ...         [0.4, 0.3, 0.3],
    ...         [0.1, 0.8, float("nan")],
    ...         [0.2, 0.3, 0.5],
    ...         [0.4, 0.4, 0.2],
    ...         [0.1, 0.2, 0.7],
    ...     ]
    ... )
    >>> preprocess_score_multiclass(y_true, y_score)
    (array([ 0.,  0.,  1.,  1.,  2., nan]),
     array([[0.7, 0.2, 0.1],
            [0.4, 0.3, 0.3],
            [0.1, 0.8, nan],
            [0.2, 0.3, 0.5],
            [0.4, 0.4, 0.2],
            [0.1, 0.2, 0.7]]))
    >>> preprocess_score_multiclass(y_true, y_score, drop_nan=True)
    (array([0., 0., 1., 2.]),
     array([[0.7, 0.2, 0.1],
            [0.4, 0.3, 0.3],
            [0.2, 0.3, 0.5],
            [0.4, 0.4, 0.2]]))

    ```
    """
    if y_true.size == 0 and y_score.size == 0:
        return np.array([]), np.array([])

    if y_true.size != y_true.shape[0]:
        msg = (
            f"'y_true' must be a an array of shape (n_samples,) or (n_samples, 1) but received: "
            f"{y_true.shape}"
        )
        raise RuntimeError(msg)
    y_true = y_true.ravel()
    if y_true.shape[0] != y_score.shape[0]:
        msg = (
            f"'y_true' and 'y_score' have different first dimension: {y_true.shape} vs "
            f"{y_score.shape}"
        )
        raise RuntimeError(msg)

    if y_score.ndim != 2:
        msg = f"'y_score' must be a 2d array but received an array of shape: {y_score.shape}"
        raise RuntimeError(msg)

    if not drop_nan:
        return y_true, y_score

    # Remove NaN values
    mask = np.logical_not(np.logical_or(np.isnan(y_true), np.isnan(y_score).any(axis=1)))
    return y_true[mask], y_score[mask]


def preprocess_score_multilabel(
    y_true: np.ndarray, y_score: np.ndarray, drop_nan: bool = False
) -> tuple[np.ndarray, np.ndarray]:
    r"""Preprocess ``y_true`` and ``y_score`` arrays for the multilabel
    classification case.

    Args:
        y_true: The ground truth target labels. This input must be an
            array of shape ``(n_samples, n_classes)`` or
            ``(n_samples,)``.
        y_score: The predicted labels. This input must be an
            array of shape ``(n_samples, n_classes)`` or
            ``(n_samples,)``.
        drop_nan: If ``True``, the NaN values are removed,
            otherwise they are kept.

    Returns:
        A tuple with the preprocessed ``y_true`` and ``y_score``
            arrays.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.utils import preprocess_score_multilabel
    >>> y_true = np.array([[1, float("nan"), 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]])
    >>> y_score = np.array(
    ...     [[2, -1, -1], [-1, 1, 2], [0, 2, 3], [3, -2, -4], [1, float("nan"), -5]]
    ... )
    >>> preprocess_score_multilabel(y_true, y_score)
    (array([[ 1., nan,  1.],
            [ 0.,  1.,  0.],
            [ 0.,  1.,  0.],
            [ 1.,  0.,  1.],
            [ 1.,  0.,  1.]]),
     array([[ 2., -1., -1.],
            [-1.,  1.,  2.],
            [ 0.,  2.,  3.],
            [ 3., -2., -4.],
            [ 1., nan, -5.]]))
    >>> preprocess_score_multilabel(y_true, y_score, drop_nan=True)
    (array([[0., 1., 0.],
            [0., 1., 0.],
            [1., 0., 1.]]),
     array([[-1.,  1.,  2.],
            [ 0.,  2.,  3.],
            [ 3., -2., -4.]]))

    ```
    """
    if y_true.size == 0 and y_score.size == 0:
        return np.array([]), np.array([])

    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))
    if y_true.ndim != 2:
        msg = f"'y_true' must be a 1d or 2d array but received an array of shape: {y_true.shape}"
        raise RuntimeError(msg)
    if y_score.ndim == 1:
        y_score = y_score.reshape((-1, 1))
    check_same_shape_score(y_true, y_score)

    if not drop_nan:
        return y_true, y_score

    # Remove NaN values
    mask = np.logical_not(
        np.logical_or(np.isnan(y_true).any(axis=1), np.isnan(y_score).any(axis=1))
    )
    return y_true[mask], y_score[mask]



r"""Implement the energy distance between two 1D distributions."""

from __future__ import annotations

__all__ = ["energy_distance"]


from typing import TYPE_CHECKING

from arkas.metric.utils import contains_nan, preprocess_same_shape_arrays
from arkas.utils.imports import check_scipy, is_scipy_available

if is_scipy_available():
    from scipy import stats

if TYPE_CHECKING:
    import numpy as np


def energy_distance(
    u_values: np.ndarray,
    v_values: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the energy distance between two 1D distributions.

    Args:
        u_values: The values observed in the (empirical) distribution.
        v_values: The values observed in the (empirical) distribution.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import energy_distance
    >>> energy_distance(u_values=np.array([1, 2, 3, 4, 5]), v_values=np.array([1, 2, 3, 4, 5]))
    {'count': 5, 'energy_distance': 0.0}

    ```
    """
    check_scipy()
    u_values, v_values = preprocess_same_shape_arrays(
        arrays=[u_values.ravel(), v_values.ravel()], drop_nan=nan_policy == "omit"
    )
    u_nan = contains_nan(arr=u_values, nan_policy=nan_policy, name="'u_values'")
    v_nan = contains_nan(arr=v_values, nan_policy=nan_policy, name="'v_values'")

    count = u_values.size
    dist = float("nan")
    if count > 0 and not u_nan and not v_nan:
        dist = float(stats.energy_distance(u_values, v_values))
    return {
        f"{prefix}count{suffix}": count,
        f"{prefix}energy_distance{suffix}": dist,
    }



r"""Contain the base class to implement a state."""

from __future__ import annotations

__all__ = ["BaseState"]

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )


class BaseState(ABC):
    r"""Define the base class to implement a state.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.state import AccuracyState
    >>> state = AccuracyState(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     y_true_name="target",
    ...     y_pred_name="pred",
    ... )
    >>> state
    AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')

    ```
    """

    @abstractmethod
    def clone(self, deep: bool = True) -> Self:
        r"""Return a copy of the state.

        Args:
            deep: If ``True``, it returns a deep copy of the state,
                otherwise it returns a shallow copy.

        Returns:
            A copy of the state.

         Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.state import AccuracyState
        >>> state = AccuracyState(
        ...     y_true=np.array([1, 0, 0, 1, 1]),
        ...     y_pred=np.array([1, 0, 0, 1, 1]),
        ...     y_true_name="target",
        ...     y_pred_name="pred",
        ... )
        ... cloned_state = state.clone()

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two states are equal or not.

        Args:
            other: The other state to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two states are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.state import AccuracyState
        >>> state1 = AccuracyState(
        ...     y_true=np.array([1, 0, 0, 1, 1]),
        ...     y_pred=np.array([1, 0, 0, 1, 1]),
        ...     y_true_name="target",
        ...     y_pred_name="pred",
        ... )
        >>> state2 = AccuracyState(
        ...     y_true=np.array([1, 0, 0, 1, 1]),
        ...     y_pred=np.array([1, 0, 0, 1, 1]),
        ...     y_true_name="target",
        ...     y_pred_name="pred",
        ... )
        >>> state3 = AccuracyState(
        ...     y_true=np.array([1, 0, 0, 0, 0]),
        ...     y_pred=np.array([1, 0, 0, 1, 1]),
        ...     y_true_name="target",
        ...     y_pred_name="pred",
        ... )
        >>> state1.equal(state2)
        True
        >>> state1.equal(state3)
        False

        ```
        """


class StateEqualityComparator(BaseEqualityComparator[BaseState]):
    r"""Implement an equality comparator for ``BaseState`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> StateEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseState, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseState):  # pragma: no cover
    EqualityTester.add_comparator(BaseState, StateEqualityComparator())



r"""Contain the implementation of a simple plotter."""

from __future__ import annotations

__all__ = ["Plotter"]

from typing import Any

from coola import objects_are_equal

from arkas.plotter.base import BasePlotter


class Plotter(BasePlotter):
    r"""Implement a simple plotter.

    Args:
        figures: The dictionary of figures.

    Example usage:

    ```pycon

    >>> from arkas.plotter import Plotter
    >>> plotter = Plotter()
    >>> plotter
    Plotter(count=0)
    >>> plotter.plot()
    {}

    ```
    """

    def __init__(self, figures: dict | None = None) -> None:
        self._figures = figures or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={len(self._figures):,})"

    def compute(self) -> Plotter:
        return Plotter(self._figures)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._figures, other._figures, equal_nan=equal_nan)

    def plot(self, prefix: str = "", suffix: str = "") -> dict:
        return {f"{prefix}{key}{suffix}": value for key, value in self._figures.items()}



r"""Contain an evaluator that sequentially evaluates several
evaluators."""

from __future__ import annotations

__all__ = ["SequentialEvaluator"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_sequence

from arkas.evaluator import BaseEvaluator, setup_evaluator
from arkas.result import EmptyResult, Result, SequentialResult

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from arkas.result import BaseResult

logger = logging.getLogger(__name__)


class SequentialEvaluator(BaseEvaluator):
    r"""Implement an evaluator that sequentially evaluates several
    evaluators.

    Args:
        evaluators: The sequence of evaluators to evaluate.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import (
    ...     SequentialEvaluator,
    ...     BinaryPrecisionEvaluator,
    ...     BinaryRecallEvaluator,
    ... )
    >>> evaluator = SequentialEvaluator(
    ...     [
    ...         BinaryPrecisionEvaluator(y_true="target", y_pred="pred"),
    ...         BinaryRecallEvaluator(y_true="target", y_pred="pred"),
    ...     ]
    ... )
    >>> evaluator
    SequentialEvaluator(
      (0): BinaryPrecisionEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
      (1): BinaryRecallEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
    )
    >>> data = pl.DataFrame({"pred": [1, 0, 0, 1, 1], "target": [1, 0, 0, 1, 1]})
    >>> result = evaluator.evaluate(data)
    >>> result
    SequentialResult(count=2)
    >>> result = evaluator.evaluate(data, lazy=False)
    >>> result
    Result(metrics=3, figures=1)

    ```
    """

    def __init__(self, evaluators: Sequence[BaseEvaluator | dict]) -> None:
        self._evaluators = tuple(setup_evaluator(evaluator) for evaluator in evaluators)

    def __repr__(self) -> str:
        args = repr_indent(repr_sequence(self._evaluators))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> BaseResult:
        out = SequentialResult(
            [evaluator.evaluate(data=data, lazy=lazy) for evaluator in self._evaluators]
        )
        if lazy or isinstance(out, EmptyResult):
            return out
        return Result(metrics=out.compute_metrics(), figures=out.generate_figures())



r"""Implement the F-beta result."""

from __future__ import annotations

__all__ = [
    "BaseFbetaScoreResult",
    "BinaryFbetaScoreResult",
    "MulticlassFbetaScoreResult",
    "MultilabelFbetaScoreResult",
]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.classification.fbeta import (
    binary_fbeta_score,
    multiclass_fbeta_score,
    multilabel_fbeta_score,
)
from arkas.metric.utils import check_nan_policy, check_same_shape_pred
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np


class BaseFbetaScoreResult(BaseResult):
    r"""Implement the base class to implement the F-beta results.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        betas: The betas used to compute the F-beta scores.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryFbetaScoreResult
    >>> result = BinaryFbetaScoreResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryFbetaScoreResult(y_true=(5,), y_pred=(5,), betas=(1,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'f1': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        betas: Sequence[float] = (1,),
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_pred = y_pred
        self._betas = tuple(betas)

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "betas": self._betas,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def betas(self) -> tuple[float, ...]:
        return self._betas

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and objects_are_equal(self.betas, other.betas, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )


class BinaryFbetaScoreResult(BaseFbetaScoreResult):
    r"""Implement the F-beta result for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with ``0`` and
            ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, *)`` with ``0`` and ``1`` values.
        betas: The betas used to compute the F-beta scores.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryFbetaScoreResult
    >>> result = BinaryFbetaScoreResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryFbetaScoreResult(y_true=(5,), y_pred=(5,), betas=(1,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'f1': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        betas: Sequence[float] = (1,),
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(
            y_true=y_true.ravel(), y_pred=y_pred.ravel(), betas=betas, nan_policy=nan_policy
        )

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return binary_fbeta_score(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            betas=self._betas,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}


class MulticlassFbetaScoreResult(BaseFbetaScoreResult):
    r"""Implement the F-beta result for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with values in
            ``{0, ..., n_classes-1}``.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, *)`` with values in
            ``{0, ..., n_classes-1}``.
        betas: The betas used to compute the F-beta scores.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MulticlassFbetaScoreResult
    >>> result = MulticlassFbetaScoreResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ... )
    >>> result
    MulticlassFbetaScoreResult(y_true=(6,), y_pred=(6,), betas=(1,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 6,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        betas: Sequence[float] = (1,),
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(
            y_true=y_true.ravel(), y_pred=y_pred.ravel(), betas=betas, nan_policy=nan_policy
        )

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multiclass_fbeta_score(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            betas=self._betas,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}


class MultilabelFbetaScoreResult(BaseFbetaScoreResult):
    r"""Implement the F-beta result for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)`` with ``0``
            and ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, n_classes)`` with ``0`` and ``1``
            values.
        betas: The betas used to compute the F-beta scores.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MultilabelFbetaScoreResult
    >>> result = MultilabelFbetaScoreResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ... )
    >>> result
    MultilabelFbetaScoreResult(y_true=(5, 3), y_pred=(5, 3), betas=(1,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5,
     'f1': array([1., 1., 1.]),
     'macro_f1': 1.0,
     'micro_f1': 1.0,
     'weighted_f1': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        betas: Sequence[float] = (1,),
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(y_true=y_true, y_pred=y_pred, betas=betas, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multilabel_fbeta_score(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            betas=self._betas,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Contain the balanced accuracy analyzer."""

from __future__ import annotations

__all__ = ["BalancedAccuracyAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from arkas.analyzer.columns import BaseTruePredAnalyzer
from arkas.metric.utils import check_nan_policy
from arkas.output.balanced_accuracy import BalancedAccuracyOutput
from arkas.state.accuracy import AccuracyState
from arkas.utils.array import to_array

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class BalancedAccuracyAnalyzer(BaseTruePredAnalyzer):
    r"""Implement the balanced accuracy analyzer.

    Args:
        y_true: The column name of the ground truth target
            labels.
        y_pred: The column name of the predicted labels.
        drop_nulls: If ``True``, the rows with null values in
            ``y_true`` or ``y_pred`` columns are dropped.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import BalancedAccuracyAnalyzer
    >>> analyzer = BalancedAccuracyAnalyzer(y_true="target", y_pred="pred")
    >>> analyzer
    BalancedAccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
    >>> frame = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> output = analyzer.analyze(frame)
    >>> output
    BalancedAccuracyOutput(
      (state): AccuracyState(y_true=(6,), y_pred=(6,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    def __init__(
        self,
        y_true: str,
        y_pred: str,
        drop_nulls: bool = True,
        missing_policy: str = "raise",
        nan_policy: str = "propagate",
    ) -> None:
        super().__init__(
            y_true=y_true, y_pred=y_pred, drop_nulls=drop_nulls, missing_policy=missing_policy
        )
        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true,
                "y_pred": self._y_pred,
                "drop_nulls": self._drop_nulls,
                "missing_policy": self._missing_policy,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def _analyze(self, frame: pl.DataFrame) -> BalancedAccuracyOutput:
        logger.info(
            f"Evaluating the balanced accuracy | y_true={self._y_true!r} | "
            f"y_pred={self._y_pred!r} | drop_nulls={self._drop_nulls} | "
            f"nan_policy={self._nan_policy!r}"
        )
        return BalancedAccuracyOutput(
            state=AccuracyState(
                y_true=to_array(frame[self._y_true]).ravel(),
                y_pred=to_array(frame[self._y_pred]).ravel(),
                y_true_name=self._y_true,
                y_pred_name=self._y_pred,
            ),
            nan_policy=self._nan_policy,
        )



r"""Contain the implementation of a figure exporter."""

from __future__ import annotations

__all__ = ["FigureExporter"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from coola.utils.path import sanitize_path
from iden.io import BaseSaver, PickleSaver, setup_saver

from arkas.exporter.base import BaseExporter

if TYPE_CHECKING:
    from pathlib import Path

    from arkas.output.base import BaseOutput

logger = logging.getLogger(__name__)


class FigureExporter(BaseExporter):
    r"""Implement a simple figure exporter.

    Args:
        path: The path where to save the figures.
        saver: The figure saver or its configuration.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            an exception is raised if the path already exists.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> import numpy as np
    >>> from arkas.output import AccuracyOutput
    >>> from arkas.state import AccuracyState
    >>> from arkas.exporter import FigureExporter
    >>> output = AccuracyOutput(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     exporter = FigureExporter(path=Path(tmpdir).joinpath("figures.pkl"))
    ...     exporter.export(output)
    ...

    ```
    """

    def __init__(
        self,
        path: Path | str,
        saver: BaseSaver | dict | None = None,
        exist_ok: bool = False,
    ) -> None:
        self._path = sanitize_path(path)
        self._saver = setup_saver(saver or PickleSaver())
        self._exist_ok = exist_ok

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "path": self._path,
                    "saver": self._saver,
                    "exist_ok": self._exist_ok,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def export(self, output: BaseOutput) -> None:
        logger.info("Exporting figures...")
        logger.info("Computing figures...")
        figures = output.get_plotter().plot()
        logger.info(f"Saving figures at {self._path}...")
        self._saver.save(figures, path=self._path, exist_ok=self._exist_ok)



r"""Contain the base class to implement an evaluator."""

from __future__ import annotations

__all__ = ["BaseEvaluator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig


class BaseEvaluator(ABC):
    r"""Define the base class to implement an evaluator.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.evaluator2 import AccuracyEvaluator
    >>> from arkas.state import AccuracyState
    >>> evaluator = AccuracyEvaluator(
    ...     AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> evaluator
    AccuracyEvaluator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    @abstractmethod
    def compute(self) -> BaseEvaluator:
        r"""Compute the metrics and return a new evaluator.

        Returns:
            A new evaluator with the computed metrics.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.evaluator2 import AccuracyEvaluator
        >>> from arkas.state import AccuracyState
        >>> evaluator = AccuracyEvaluator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> evaluator2 = evaluator.compute()
        >>> evaluator2
        Evaluator(count=5)

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two evaluators are equal or not.

        Args:
            other: The other evaluator to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two evaluators are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.evaluator2 import AccuracyEvaluator
        >>> from arkas.state import AccuracyState
        >>> evaluator1 = AccuracyEvaluator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> evaluator2 = AccuracyEvaluator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> evaluator3 = AccuracyEvaluator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 0, 0]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> evaluator1.equal(evaluator2)
        True
        >>> evaluator1.equal(evaluator3)
        False

        ```
        """

    @abstractmethod
    def evaluate(self, prefix: str = "", suffix: str = "") -> dict:
        r"""Evaluate the metrics.

        Args:
            prefix: The key prefix in the returned dictionary.
            suffix: The key suffix in the returned dictionary.

        Returns:
            The metrics.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.evaluator2 import AccuracyEvaluator
        >>> from arkas.state import AccuracyState
        >>> evaluator = AccuracyEvaluator(
        ...     AccuracyState(
        ...         y_true=np.array([1, 0, 0, 1, 1]),
        ...         y_pred=np.array([1, 0, 0, 1, 1]),
        ...         y_true_name="target",
        ...         y_pred_name="pred",
        ...     )
        ... )
        >>> evaluator.evaluate()
        {'accuracy': 1.0, 'count_correct': 5, 'count_incorrect': 0, 'count': 5, 'error': 0.0}

        ```
        """


class EvaluatorEqualityComparator(BaseEqualityComparator[BaseEvaluator]):
    r"""Implement an equality comparator for ``BaseEvaluator``
    objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> EvaluatorEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseEvaluator, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseEvaluator):  # pragma: no cover
    EqualityTester.add_comparator(BaseEvaluator, EvaluatorEqualityComparator())



r"""Contain utility functions to generate HTML reports."""

from __future__ import annotations

__all__ = ["create_html_report"]

from coola.utils import str_indent


def create_html_report(toc: str, body: str) -> str:
    r"""Create a HTML report.

    Args:
        toc: The table of contents of the report.
        body: The body of the report

    Returns:
        The HTML report.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.reporter.utils import create_html_report
    >>> report = create_html_report(toc="", body="custom HTML with the content of the report")

    ```
    """
    return f"""
<!doctype html>
<html>
    <head>
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.1/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-4bw+/aepP/YC94hEpVNVgiZdgIC5+VKNBQNGCHeKRQN+PtmoHDEXuppvnDJzQIu9" crossorigin="anonymous>"
    </head>
    <body style="margin: 1.5rem;">
    <div id="toc_container">
    <h2>Table of content</h2>
    {str_indent(toc, num_spaces=4)}
    </div>

    \t{str_indent(body, num_spaces=8)}
    </body>
</html>
"""
